/*  main.cpp
	mp3player on sc12

	15.05.2001: tk, initial implementation.
	04.01.2002: ab, splitt frontend and backend (now playlist.cpp)
	12.01.2002: ab, add support subdirectories
	15.01.2002: ab, redesign "ir-kbhit" with RC_ENUM
	05.02.2002: ab, support load/store config
	10.02.2002: ab, add repeat funktion
	12.02.2002: ab, add random seek in playlist
	22.02.2002: ab, change Keys from RC_ENUM to VK
	23.02.2002: ab, remove some static char buffers

	02.03.2002: ab, add playmode repeat/random only directory
	13.03.2002: ab, add telnet/console support

	Copyright (c)2000 by Thomas Kindler, thomas.kindler@gmx.de
	2002    (ab)Andreas Brachold, a.brachold@berlin.de

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License as
	published by the Free Software Foundation; either version 2 of
	the License, or (at your option) any later version. Read the
	full License at http://www.gnu.org/copyleft for more details.
*/

// include files -----
//
#include <alloc.h>
#include <stdio.h>
#include <dos.h>
#include <dir.h>
#include <io.h>
#include <time.h>
#include <conio.h>
#include <stdlib.h>

#include "checkbios.h"
#include "vs1001.h"
#include "inet/httpget.h"
#include "inet/url.h"
#include "mpbuf.h"
#include "misc.h"
#include "remote.h"
#include "mpeg.h"

#include "main.h"
#include "playlist.h"
#include "vfd.h"


/// current display luminance
int Luminance = 0;
/// current speaker volume
int Volume = 0xC0;
/// current speaker volume
unsigned int nPlayFile = 0;
/// current speaker bassmode
bool bModeBass = true;
/// current speaker mutemode
bool bModeMute = false;

/// current playmode
int nPlayMode = PLAY; //Play -> Repeat -> Random -> Repeat Dir -> Random Dir

/// current displaymode
int nDisplayInfo = 0;
/// current keystroke layer. false - layer 0, true - layer 1
bool bModeShift = false;

/// count founded files
unsigned int nFiles = 0;

/// Filename for the configuration
static const char *iniFile = "a:\\mp3.cfg";

/// load settings from cfg file.
void loadsettings()
{
    FILE *f = fopen(iniFile, "rb");
    if(f != NULL)
    {
	if(!feof(f)) Volume = fgetc(f);
	if(!feof(f)) Luminance = fgetc(f);
	if(!feof(f)) nPlayFile = fgetc(f);
	if(!feof(f)) bModeBass = 0 != fgetc(f);
	if(!feof(f)) bModeMute = 0 != fgetc(f);
	if(!feof(f)) nPlayMode = fgetc(f);
	fclose(f);
    }
}
/// save current settings to cfg file.
void savesettings()
{
    FILE *f = fopen(iniFile, "wb");
    if(f != NULL)
    {
   	if(EOF == fputc( Volume,f)
	    || EOF == fputc( Luminance,f)
	    || EOF == fputc( nPlayFile,f)
	    || EOF == fputc( bModeBass == true,f)
	    || EOF == fputc( bModeMute == true,f)
	    || EOF == fputc( nPlayMode,f))
	    {
		printf("failed savesettings\n");
	    };
	fclose(f);
    }
}

int main(int argc, char *argv[])
{
	// start search from directory e.g. b:\musik
	char directory[MAXPATH] = "b:\\";

    int nVK_Pressed;

    if (!CheckBIOS())
	return -1;

    SYS_SetSTDIOFocus(SYS_FOCUS_APP);
    printf("k/os mp3 player v2.0 pre-alpha\n"
		"(c)1999-2001 by thomas.kindler@gmx.de.\n"
		"   2002 by a.brachold@berlin.de.\n"
		"Build %s %s\n", __DATE__, __TIME__);

    loadsettings();

    VFD_Init(Luminance);
    RC_Init();

    BUF_Init(16 * 1024, 128 * 1024L);

    VS_Init();
    VS_SetVolume(Volume, Volume);

    printf("\n");

    if (argc == 3)
    {
		if (0 == strcmpi(argv[1], "-f"))
		{
	    //3nd paramter is a single mp3 file
	    play(argv[2]);
        }
		else
		{
	    //3nd paramter is URL
	    http_get(argv[2]);
        }
    }
    else
    {
   		if (argc == 2) //2nd paramter is startdirectory
	    	strncpy(directory, argv[1], sizeof(directory));

        randomize();

		nFiles = filelist_get(directory);
		//printf("found %d files\n", nFiles);

		if (nFiles)
	    {
	    	// avoid invalid cfg files;
			nPlayMode %= LASTPLAYMODE;

		    if(nPlayMode == RANDOM ) //init random filenummer
	  	    {
			nPlayFile = rand() % nFiles;
		    }
    	    else
		    {
	        	if (nPlayFile >= nFiles) // Limit last played file
		            nPlayFile = 0;
    	    }

		    // the endloop on founded files
		    while (VK_POWER != (nVK_Pressed = playfile(nPlayFile,nPlayMode)))
            {
        	// Check last user-rc-code
			if (nVK_Pressed == VK_NEXT | nVK_Pressed == VK_PREV | nVK_Pressed == VK_PLAY )
        	{
            	nPlayFile =  getnextfile(nPlayFile,nFiles,nPlayMode,nVK_Pressed != VK_PREV);
			}//if(nRet== VK..
	    }//while
	}
//if(nFiles)
    }//if(argc==
    
    savesettings();
    filelist_free();

    BUF_Done();
    VS_Done();

    RC_Done();
    VFD_Clear();
    VFD_Done();

    SYS_SetSTDIOFocus(SYS_FOCUS_BOTH);
    printf("\ndone...\n");

    return 0;
}


/// Helpermacro for switch(RC_GetKey())
#define CHECK_KEY(RC5CODE,EMUMCODE)	\
	case(RCKEY_RC5 | RC5CODE):	\
         {                             	\
           	nRet = EMUMCODE;	\
	         break;			\
         }
/// Helpermacro for switch(RC_GetKey())
#define CHECK_KEYS(RC5CODE,EMUMCODE1,EMUMCODE2) 	\
	case(RCKEY_RC5 | RC5CODE):						\
        {                                   		\
	    if(bModeShift)	nRet = (EMUMCODE2);			\
	    else		nRet = (EMUMCODE1);  			\
      	    break;                              	\
        }

/// Helpermacro for switch(getch())
#define CHECK_KB(KEY,EMUMCODE)		  				\
	case(KEY):										\
    {                             					\
    	nRet = EMUMCODE;							\
	         break;									\
    }

/**
TRANSLATE RC Codec to Usercodec
*/
int remote_get(int rc_max)
{
    int nRet = VK_EMPTY;
    // Process user input
    //
    while (RC_KeyAvail())
    {
	unsigned long key = RC_GetKey();
	switch (key & (RCKEY_PROTO | RCKEY_CODE))
    	{
	    // RC Code den Virtual Key
	    CHECK_KEY( 0x68c ,VK_POWER);// Taste Power
	    CHECK_KEY( 0x68f ,VK_INFO);// Taste Info
	    CHECK_KEY( 0x6b0 ,VK_PAUSE);// Taste Pause
	    CHECK_KEY( 0x6b2 ,VK_FAST_REWIND);// Taste <<
	    CHECK_KEY( 0x6b4 ,VK_FAST_FORWARD);// Taste >>
	    CHECK_KEY( 0x6b5 ,VK_PLAY);// Taste Play
	    CHECK_KEY( 0x6b6 ,VK_STOP);// Taste Stop
	    CHECK_KEY( 0x6a0 ,VK_NEXT);// Taste Channel++
	    CHECK_KEY( 0x6a1 ,VK_PREV);// Taste Channel--

	    // Doublelayer keys (without and with shift)
	    CHECK_KEY( 0x7fe ,VK_SHIFT);// Taste Shift
        CHECK_KEYS( 0x410 ,VK_VOLUME_UP,VK_LUM_UP);// Taste Vol++
        CHECK_KEYS( 0x411 ,VK_VOLUME_DOWN,VK_LUM_DOWN);// Taste Vol--
        CHECK_KEYS( 0x40d ,VK_VOLUME_MUTE,VK_VOLUME_BASS);// Taste Mute

	     // TODO keys
	    CHECK_KEY( 0x6b7 ,VK_DELETE);// Taste Record
	    CHECK_KEY( 0x680 ,VK_0);// Taste 0
	    CHECK_KEY( 0x681 ,VK_1);// Taste 1
	    CHECK_KEY( 0x682 ,VK_2);// Taste 2
	    CHECK_KEY( 0x683 ,VK_3);// Taste 3
        CHECK_KEY( 0x684 ,VK_4);// Taste 4
	    CHECK_KEY( 0x685 ,VK_5);// Taste 5
	    CHECK_KEY( 0x686 ,VK_6);// Taste 6
	    CHECK_KEY( 0x687 ,VK_7);// Taste 7
	    CHECK_KEY( 0x688 ,VK_8);// Taste 8
	    CHECK_KEY( 0x689 ,VK_9);// Taste 9

	default:
	    printf("\runknown RC code: 0x%lx\n", key);
	    break;
	}
	}
	// No IR Code detected then check console keyboard
	if(VK_EMPTY == nRet)
    while(kbhit())
    {
       char taste = getch();
       switch(taste)
       {
	    // RC Code den Virtual Key
	    CHECK_KB( 'q' ,VK_POWER);// Taste Power
	    CHECK_KB( 'i' ,VK_INFO);// Taste Info
	    CHECK_KB( 'w' ,VK_PAUSE);// Taste Pause
	    CHECK_KB( 'r' ,VK_FAST_REWIND);// Taste <<
	    CHECK_KB( 'f' ,VK_FAST_FORWARD);// Taste >>
	    CHECK_KB( ' ' ,VK_PLAY);// Taste Play
	    CHECK_KB( 's' ,VK_STOP);// Taste Stop
	    CHECK_KB( 'n' ,VK_NEXT);// Taste Channel++
	    CHECK_KB( 'p' ,VK_PREV);// Taste Channel--

	    // Doublelayer keys (without and with shift)
        CHECK_KB( '+' , VK_VOLUME_UP );
        CHECK_KB( '-' , VK_VOLUME_DOWN);
        CHECK_KB( 'l' , VK_LUM_UP);// Taste Vol++
        CHECK_KB( 'd' , VK_LUM_DOWN);// Taste Vol--
        CHECK_KB( 'm' , VK_VOLUME_MUTE)
        CHECK_KB( 'b' , VK_VOLUME_BASS);// Taste Mute

	     // TODO keys
	    CHECK_KB( 'x' ,VK_DELETE);// Taste
	    CHECK_KB( '0' ,VK_0);// Taste 0
	    CHECK_KB( '1' ,VK_1);// Taste 1
	    CHECK_KB( '2' ,VK_2);// Taste 2
	    CHECK_KB( '3' ,VK_3);// Taste 3
        CHECK_KB( '4' ,VK_4);// Taste 4
	    CHECK_KB( '5' ,VK_5);// Taste 5
	    CHECK_KB( '6' ,VK_6);// Taste 6
	    CHECK_KB( '7' ,VK_7);// Taste 7
	    CHECK_KB( '8' ,VK_8);// Taste 8
	    CHECK_KB( '9' ,VK_9);// Taste 9
		default:
		    printf("\runknown key: 0x%x\n", taste);
	    	break;
       }
    }
	 // Check Limit RCCODE ( for e.g. "streams can't seek forward")
    if(nRet <= rc_max)
    	return nRet;
    return VK_EMPTY;
}

// defined Text for display Status
static const char *dslPwr = "Beende...";
static const char *dslOpe = "ffne...";
static const char *dslPly = "Spiele...";
static const char *dslPse = "Pause...";
static const char *dslStp = "Stoppe...";
// defined Text for display Togglefunktions
static const char *dslOn  = "ein";
static const char *dslOff = "aus";
static const char *dslGrp = "nur Gruppe ";
static const char *dslSft = "Shift ";
static const char *dslRpt = "Wiederhole ";
static const char *dslRdm = "Zufall ";
static const char *dslMte = "Stumm ";
static const char *dslBss = "Bass ";
// defined Text for display Progress (Only 3 Character)
static const char *dslVol = "Vol";
static const char *dslLum = "Lum";

static const char *dslFwd = "\xeb\xeb\x20";//">> ";
static const char *dslRwd = "\xec\xec\x20";//"<< ";
static const char *dslNxt = "\xeb| ";//">| ";
static const char *dslPrv = "|\xec ";//"|< ";
static const char *dslHlt = " \x7F ";//" [] ";

int display(const char *szFirstLine, long xPos, long xMax, int rc_max)
{
    // value for delay counter wait before change direction
    const int nDelayLine = 20;
    // Length for firstline without scrolling
    const int nScrollLength = 20;
    // delay counter for scroll firstline
    static int n1Scroll = {-nDelayLine};
    // remember current scrolldirection
    static int n1ScrollDir = {1};
    // counter for delay key - toggle
    static int nToggleDelay = {0};
    // delay counter switch back for secondline
    static int n2DelaySwitchBack = {0};

    static char szSecondLine[80];

    int key = remote_get(rc_max);
    if (key != VK_EMPTY)// Is press any key ?
    {
	// Display delay counter
       	n2DelaySwitchBack = 15;

	/************************************************
	* keystroke Power */
	if (key == VK_POWER)
      	{
	    BUF_Clear();
	    n1Scroll = -nDelayLine;
	    strncpy(szSecondLine, dslPwr, sizeof(szSecondLine));
	}
	/************************************************
	* keystroke play */
	if (key == VK_PLAY && nToggleDelay == 0)
      	{
    	    if(!VS_IsPlaying())
    	    {
			n1Scroll = -nDelayLine;
			strncpy(szSecondLine, dslPly, sizeof(szSecondLine));
	    	}
      	    else
	    	{
				// rotate Funktion 1.Play -> 2.Repeat -> 3.Random
				++nPlayMode %= LASTPLAYMODE;
                switch( nPlayMode)
                {
		    	case REPEATGROUP:
                case REPEAT:
	            {
					strncpy(szSecondLine, dslRpt, sizeof(szSecondLine));
                    if(nPlayMode == REPEATGROUP)
                    	strncat(szSecondLine, dslGrp,  sizeof(szSecondLine));
					strncat(szSecondLine, dslOn,  sizeof(szSecondLine));
                    break;
		    	}
		    	case RANDOMGROUP:
		    	case RANDOM:
	            {
					strncpy(szSecondLine, dslRdm, sizeof(szSecondLine));
                    if(nPlayMode == RANDOMGROUP)
                    	strncat(szSecondLine, dslGrp,  sizeof(szSecondLine));
					strncat(szSecondLine, dslOn,  sizeof(szSecondLine));
                    break;
		    	}
		    	case PLAY:
	            {
					strncpy(szSecondLine, dslPly, sizeof(szSecondLine));
                	break;
			    }
			}
         	key = VK_EMPTY; // Delete follow keys ( first key is valid)
	    }
	}
	/************************************************
	* Pause */
	if (key == VK_PAUSE)
    {
	   	if(VS_IsPlaying())
       	{
			strncpy(szSecondLine, dslPse, sizeof(szSecondLine));
		   }
      	    else key = VK_EMPTY; // Delete follow keys ( first key is valid)
	}
   	if (key == VK_STOP)
    {
	    if(VS_IsPlaying())
	    {
			strncpy(szSecondLine, dslStp, sizeof(szSecondLine));
	    }
		else key = VK_EMPTY; // Delete follow keys ( first key is valid)
    }

	/************************************************
	* Shift */
	if (key == VK_SHIFT && nToggleDelay == 0)  // Only first key
    {
	   	bModeShift = !bModeShift;

        strncpy(szSecondLine, dslSft, sizeof(szSecondLine));
	    strncat(szSecondLine, bModeShift ? dslOn : dslOff,
                sizeof(szSecondLine));
	}
	/************************************************
	* Info */
	if (key == VK_INFO && nToggleDelay == 0) // Only first key
    {
	   	n2DelaySwitchBack = 0; // delay don't use her
		// rotate Funktion 1.MPEG -> 2.Starttime -> 3.Endtime
	   	++nDisplayInfo %= LASTDISPLAYINFO;
	}

	/************************************************
	* Bass */
	if (key == VK_VOLUME_BASS && nToggleDelay == 0) // Only first key
    {
	    bModeBass = !bModeBass;
	    VS_SetBassBoost(bModeBass);

	    strncpy(szSecondLine, dslBss, sizeof(szSecondLine));
	    strncat(szSecondLine,bModeBass ? dslOn : dslOff,
		sizeof(szSecondLine));
	}
	/************************************************
	* Mute */
	if (key == VK_VOLUME_MUTE && nToggleDelay == 0)  // Only first keystroke
	{
	    bModeMute = !bModeMute;
	    VS_SetVolume(bModeMute?129:Volume,bModeMute?129:Volume); // 128 -> VS1001 goto Standby
	    strncpy(szSecondLine, dslMte, sizeof(szSecondLine));
	    strncat(szSecondLine,bModeMute ? dslOn : dslOff, sizeof(szSecondLine));
	}
	/************************************************
	* Volume */
	if (key == VK_VOLUME_DOWN || key == VK_VOLUME_UP)
      	{
	    Volume += (key == VK_VOLUME_DOWN)?-4:4;
	    if (Volume < 129) Volume = 129;  // 128 -> VS1001 goto Standby
	    if (Volume > 255) Volume = 255;
	    VS_SetVolume(Volume, Volume);

		vfd_progressBar(szSecondLine,sizeof(szSecondLine),dslVol, Volume, 128, 255);
	}
	/************************************************
	* Luminance */
	if (key == VK_LUM_DOWN || key == VK_LUM_UP)
      	{
	    Luminance += (key == VK_LUM_DOWN)?-8:8;
	    if (Luminance < 0) 	Luminance = 0;
	    if (Luminance > 255)Luminance = 255;
	    VFD_Luminance(Luminance);

    	vfd_progressBar(szSecondLine,sizeof(szSecondLine),dslLum, Luminance, 0, 255);
	}
	/************************************************
	* Fast Forward and Rewind */
	if (key == VK_FAST_FORWARD || key == VK_FAST_REWIND)
      	{
    	    n2DelaySwitchBack = 1; // only short delay
	    vfd_progressBar(szSecondLine,sizeof(szSecondLine),key == VK_FAST_FORWARD?dslFwd:dslRwd,
				    xPos, 0, xMax);
	}
	/************************************************
	* Skip Forward and Prevorius */
	if(key == VK_NEXT  || key == VK_PREV)
   	{
	    if (nToggleDelay == 0) // Only first key
	    {
		n1Scroll = -nDelayLine;  //Reset scrollpos to start
		vfd_progressBar(szSecondLine,sizeof(szSecondLine),key == VK_NEXT?dslNxt:dslPrv,
                    					nPlayFile, 0, nFiles);
	    }
	    else key = VK_EMPTY; // Delete follow keys ( first key is valid)
	}

	// Key toggle counter
	nToggleDelay = 3;
    }// End of keydetection

    // decrement Key toggle counter from 3 to 0
    if (nToggleDelay > 0)
    	nToggleDelay--;


    if (n2DelaySwitchBack == 0) // display normal progress (overwrite szSecondLine )
    {
	unsigned long header;
	VS_GetMPEGHeader(&header);
	FrameInfo info;
	if (MPEG_CheckHeader(header)
	    && MPEG_ParseHeader(header, &info))
	{
	    switch(nDisplayInfo)
        {
        	case DISPLAYMPEG: //view header information
		    {
            	MPEG_HeaderToString(header,szSecondLine,sizeof(szSecondLine));
                break;
    	    }
			case DISPLAYRUNTIME:
            case DISPLAYENDTIME:
      	    {
		        char szBuf[17];
		        char szTime[8];
                long s,m;
	        	if(nDisplayInfo == DISPLAYENDTIME) // view time to end
            	{	//calc total seconds to end
				    s = ((xMax - xPos) / 128) / (info.bitrate/1000);
            	    *szTime     = '-';
            	}
            	else	// view time from begin
            	{	//calc total seconds from begin
				    s = (xPos / 128) / (info.bitrate/1000);
            	    *szTime     = ' '; //+
            	}

            	*(szTime+1) = '\0';

            	m = s / 60;	//calc minute
            	s = s % 60; //calc restseconds
				if(m > 99) m = 99;
				if(m < 0) m = 0;
				if(s < 0) s = 0;

				if(m < 10) // Add leading Zero
				    strncat(szTime,"0",sizeof(szTime));
	    	    strncat(szTime,itoa(m,szBuf,10),sizeof(szTime));
	        	strncat(szTime,":",sizeof(szTime));

		   		if(s < 10) // Add leading Zero
           	    	strncat(szTime,"0",sizeof(szTime));
	       		strncat(szTime,itoa(s,szBuf,10),sizeof(szTime));

				vfd_progressBar(szSecondLine,sizeof(szSecondLine),NULL, xPos, 0, xMax,szTime);
                break;
	            }
			}//switch(displayinfo)
		}
		else //why can't read valid mpeg header ?
    	{
		   	vfd_progressBar(szSecondLine,sizeof(szSecondLine),dslHlt, xPos, 0, xMax);
		}
    }
    else
    {
    	// decrement counter for backswitch 2nd Line
		n2DelaySwitchBack--;
    }
    //////////////////////////////////////////////////////////////
    // Scroll service 1st line
    int n1Len = strlen(szFirstLine);
    n1Scroll += n1ScrollDir;

    if (n1Scroll > ((n1Len - nScrollLength) * 4) + nDelayLine)
    {
	n1ScrollDir = -1;
	n1Scroll = (((n1Len - nScrollLength) * 4) + nDelayLine);
    }
    if (n1Scroll < -nDelayLine)
    {
	n1ScrollDir = 1;
	n1Scroll = -nDelayLine;
    }
    int off1 = n1Scroll / 4;
    if (off1 > n1Len - nScrollLength)
	off1 = n1Len - nScrollLength;
    if (off1 < 0)
	off1 = 0;

    /////////////////////////////////////////////////////////
    // Her is the only posistion where dumped out to display
    {
    VFD_Clear();
    vfdtext(szFirstLine + off1);
    vfdmoveto(0, 1);
    vfdtext(szSecondLine);
    VFD_Update();
	}

    /////////////////////////////////////////////////////////
    // Her is the only posistion where dumped out to console
    {
		int i;
		char sz1[21];
        char sz2[21];

        strncpy(sz1,szFirstLine + off1,sizeof(sz1));
      	for(i = strlen(sz1);i < (sizeof(sz1));++i)
        	sz1[i] = ' ';
		sz1[sizeof(sz1)-1]= '\0';
    	for(i = 0;i < (sizeof(sz2)-1) && szSecondLine[i];++i)
			if(szSecondLine[i]>=0x20)
		          sz2[i] = szSecondLine[i];
            else
            	sz2[i] = '>';
		sz2[i]= '\0';
      	for(i = strlen(sz2);i < (sizeof(sz2));++i)
        	sz2[i] = ' ';
		sz2[sizeof(sz2)-1]= '\0';
		printf("\r| %s | %s |",sz1,sz2);
    }
    return key;
}
