/*  vfd.cpp
*  Low level VFD driver for NEC 20x2 Textdisplay
*
*  01.01.2002: ab, initial implementation.
*  05.02.2002: ab, translationtable
*  12.02.2002: ab, remove floats in ProgressBar through some problems with float on SC12
*  20.02.2002: ab, update display only if changed
*  23.02.2002: ab, disable output ifn't successfull initialize
*
*   Copyright (c)2002 by Andreas Brachold, a.brachold@berlin.de
*
*    This program is free software; you can redistribute it and/or
*    modify it under the terms of the GNU General Public License as
*    published by the Free Software Foundation; either version 2 of
*    the License, or (at your option) any later version. Read the
*    full License at http://www.gnu.org/copyleft for more details.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sc12/hwapi.h"
#include "sc12/rtos.h"

#include "vfd.h"

// Width and height of the display
// 20x2 Character
static const unsigned int VFD_WIDTH = 20;
static const unsigned int VFD_HEIGHT = 2;
// PIO Pin for BUSY Signal
static const unsigned int VFD_BUSY_PIO = 7;

// memory buffer for the VFD screen.
//
static unsigned char vfd_screen[VFD_HEIGHT*VFD_WIDTH];

static int xPos,yPos;
/// is display successfull initialize
static bool bVfdSucessInit = false;

///////////////////////////////////////////////////////////////////////
//  Funktionsname	: outvfd
/** Send a single byte to the display
@return bool - true byte sended to display.
- false display disconnected or to slow
@param	unsigned char b  - true if Byte written.*/
static bool outvfd(unsigned char b)
{
    short maxDelay = 10; // Maximum wait time 10 ms

    while(hal_read_pio(VFD_BUSY_PIO)  // Wait until VFD is ready for next Character
	    && 0 != maxDelay-- )      // On Datasheet is Busytime ~ 750s
    {
	RTX_Sleep_Time(1);
    }
    	
    if(0 < maxDelay)
    {
	outp(0x500, b);  // Output byte to vfd
	return true;
    }
    //timeout busy is'nt go low
    return false;
}

static const char translate[256] =
{
    	0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
		0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,

		0x20,0x21,0x22,0x23,0x24,0x25,0x26,0x27,0x28,0x29,0x2A,0x2B,0x2C,0x2D,0x2E,0x2F,
		0x30,0x31,0x32,0x33,0x34,0x35,0x36,0x37,0x38,0x39,0x3A,0x3B,0x3C,0x3D,0x3E,0x3F,
		0x40,0x41,0x42,0x43,0x44,0x45,0x46,0x47,0x48,0x49,0x4A,0x4B,0x4C,0x4D,0x4E,0x4F,
		0x50,0x51,0x52,0x53,0x54,0x55,0x56,0x57,0x58,0x59,0x5A,0x5B,0x5C,0x5D,0x5E,0x5F,
		0x60,0x61,0x62,0x63,0x64,0x65,0x66,0x67,0x68,0x69,0x6A,0x6B,0x6C,0x6D,0x6E,0x6F,
		0x70,0x71,0x72,0x73,0x74,0x75,0x76,0x77,0x78,0x79,0x7A,0x7B,0x7C,0x7D,0x7E,0x7F,

		0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x8A,0x8B,0x8C,0x8D,0x8E,0x8F,
		0x90,0x91,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0x9A,0x9B,0x9C,0x9D,0x9E,0x9F,
		0xA0,0xA1,0xA2,0xA3,0xA4,0xA5,0xA6,0xA7,0xA8,0xA9,0xAA,0xAB,0xAC,0xAD,0xAE,0xAF,
		0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xB8,0xB9,0xBA,0xBB,0xBC,0xBD,0xBE,0xBF,
		0xC0,0xC1,0xC2,0xC3,'\xA2',0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,
		0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,'\xA3',0xD7,0xD8,0xD9,0xDA,0xDB,'\xA4',0xDD,0xDE,0xDF,
		0xE0,0xE1,0xE2,0xE3,'\xA5','\xE4',0xE6,0xE7,0xE8,0xE9,0xEA,0xEB,0xEC,0xED,0xEE,0xEF,
		0xF0,0xF1,0xF2,0xF3,0xF4,0xF5,'\xA6',0xF7,0xF8,0xF9,0xFA,0xFB,'\xA7',0xFD,0xFE,0xFF
};


///////////////////////////////////////////////////////////////////////
//  Funktionsname	: outvfdasc
/** Send a single byte to the display and translate it to locale asctable
@return bool
@param	unsigned char b  - true if Character written.*/
static bool outvfdasc(unsigned char b)
{
    return outvfd(translate[b]);
}

static bool bVFDUpdate = true;
///////////////////////////////////////////////////////////////////////
//  Funktionsname	: VFD_Update
/**  Send the screen buffer to the actual display memory
@return 	bool - true if all Character written.*/
void VFD_Update()
{
    
    if(!bVFDUpdate || !bVfdSucessInit) // Allready uptodate
	return;
	
    if(outvfd(0x0d)) //Cleardisplay
    {
		unsigned int x,y;
		for (y=0; y<VFD_HEIGHT; ++y)
	    {
			for (x=0; x<VFD_WIDTH && outvfdasc(vfd_screen[x + y*VFD_WIDTH]); ++x);
	    }
        bVFDUpdate = false;
	    //all Character written.
    }
}


///////////////////////////////////////////////////////////////////////
//  Funktionsname	: VFD_Luminance
/** Set display luminance.
@return 	bool
@param	int level - luminance level 0..255
*/
bool VFD_Luminance(int level)
{
    if(bVfdSucessInit)
    {
	switch((level >> 6) & 0x03)	//Reduce value to 2bit
	{
	   case 0x00:	return outvfd(0x04); // Set Luminance to 25%
	   case 0x01:	return outvfd(0x03); // Set Luminance to 50%
	   case 0x02:	return outvfd(0x02); // Set Luminance to 75%
	   case 0x03:	return outvfd(0x01); // Set Luminance to 100%
	}
    }
    return false;
}

///////////////////////////////////////////////////////////////////////
//  Funktionsname	: VFD_Clear
/** Clear the vfd_screen array.
@return void */
void VFD_Clear()
{
    memset(vfd_screen, 0x20, sizeof(vfd_screen));
    xPos = 0;
    yPos = 0;
    bVFDUpdate = true;
}

///////////////////////////////////////////////////////////////////////
//  Funktionsname	: VFD_Init
/** VFD_Init() must be called, before using any of the other
VFD functions. The display is initialized, and cleared.
@return 	bool - true init sucessfull
@param	int Luminance
*/
bool VFD_Init(int Luminance)
{
	printf("Init VFD Display ");

    pfe_enable_bus(0xff, true);
    pfe_enable_pcs(5);
    pfe_enable_pcs(6);
    pfe_enable_pio(VFD_BUSY_PIO,PIO_I);	// Busy as Input

    // Is display connected ?
    bVfdSucessInit = outvfd(0x0c)  // Cleardisplay
        	  && outvfd(0x11)  // Normalwrite
        	  && outvfd(0x14)  // Cursoroff
        	  && outvfd(0x18);  // Cursoroff
		  
    if(bVfdSucessInit) //preinit ok ?
    {
	bVfdSucessInit = VFD_Luminance(Luminance);

        VFD_Clear();
        VFD_Update();

	if(bVfdSucessInit)
	{
    	    printf("ok\n");
    	    return true;
	}
    }
    printf(" failed\n");
    return false;
}



///////////////////////////////////////////////////////////////////////
//  Funktionsname	: VFD_Done
/** Shut down the VFD controller.
@return 	void*/
void VFD_Done()
{
    if(bVfdSucessInit) //init ok ?
    {
	outvfd(0x0c);  // Cleardisplay
    }
    // free pio pins
    //
    pfe_enable_pio(VFD_BUSY_PIO,  PIO_IPU );

    pfe_enable_pio(13, PIO_IPD);
    pfe_enable_pio( 2, PIO_IPU);
    pfe_enable_pio( 3, PIO_IPU);
}


///////////////////////////////////////////////////////////////////////
//  Funktionsname	: vfdtext
/** Print a plain text string.
The cursor position is advanced to the end of the string.
@return void
@param	const char *text
*/
void vfdtext(const char *text)
{
	int n = VFD_WIDTH - xPos ;

	if(xPos < VFD_WIDTH && yPos < VFD_HEIGHT && n > 0)
	{
		int nlen = strlen(text);
		if(nlen < n) n = nlen;
		// Check if Displaymemory is changed
		if(0 != memcmp(vfd_screen + xPos + (yPos*VFD_WIDTH),text,n))
		{
			memcpy(vfd_screen + xPos + (yPos*VFD_WIDTH),text,n);
			bVFDUpdate = true;
		}
		xPos += n;
	}
}


///////////////////////////////////////////////////////////////////////
//  Funktionsname	: vfdprintf
/** Print a formatted text string.
The cursor position is advanced to the end of the string.
@return 	int - number of characters printed.
@param	const char *format - see printf documentation.
@param	...  - see printf documentation.*/
/*
int vfdprintf(const char *format, ...)
{
	static char buf[1024];

	va_list arglist;
	va_start(arglist, format);

	int ret = vsprintf(buf, format, arglist);
	assert(ret < sizeof(buf));

	vfdtext(buf);
	va_end(arglist);
	return ret;
}
*/

///////////////////////////////////////////////////////////////////////
//  Funktionsname	: vfdmoveto
/** Set cursor position.
@return void
@param	int x
@param	int y */

void vfdmoveto(int x,int y)
{
	xPos = x;
	yPos = y;
}

///////////////////////////////////////////////////////////////////////
//  Funktionsname	: vfd_progressBar
/**
@return 	const char*
@param	char* buf
@param	size_t buflen
@param	const char* head
@param	long v - current value
@param	long min -  minimum value
@param	long max -  maximum value */
void vfd_progressBar(char* buf,unsigned int buflen,
			    const char* head,
			    long v,long min,long max,
                            const char* szTime)
{
    char* szWheel = "   ";
    long progress;

    if(!head) // No Head defined -> display spinctrl    
	{
    // sequence for rotating wheel 
	const static char wheel[] = {"\xeb\xeb\xeb\xeb\xeb\xeb\xeb\xeb\xeb\xed\xec\xee"};
    // current position for rthe wheel
	static int  spin;
	*(szWheel + 1) = wheel[spin++ % (sizeof(wheel)-1)];
	head = szWheel;
    }
	strncpy(buf,head,buflen);	//
	strncat(buf,"__________",buflen);

    // calc optimal number between 0 - 100
    while((max-min) < 100) { max *= 10;min *= 10;v *= 10; };
    while((max-min) > 100){ max /= 10;min /= 10;v /= 10; };
	// transform from v & min & max to precent value
	if(max-min != 0)
	progress = (((v-min)*100)/((max-min)));
    if (progress<=0) progress = 1;
    if (progress>100) progress = 100;

    if(szTime) 
    {
		strncat(buf,szTime,buflen);
    }
    else 
    {
		// ifn't szTime defined display progress in precent e.g. "50%"
		strncat(buf," ",buflen);
        itoa(progress,buf + strlen(buf),10);
		strncat(buf,"%",buflen);
    }

    int n = strlen(head);

    int i=0;
    //place up to teen step for progress bar 
    for (; i<progress/10; i++) buf[i+n] = 0x7F; // 0x7F is a '[]'
    //replace last step 
    buf[i+n] = 0x87; // 0x87 is a '>'
}

